/*
**  AddressBook.m
**
**  Copyright (c) 2001, 2002, 2003
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**          Jonathan B. Leffert <jonathan@leffert.net>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AddressBook.h"

#ifdef MACOSX
#import <AddressBook/AddressBook.h>
#endif


#include "Address.h"
#include "Group.h"
#include "Constants.h"
#include "Utilities.h"

static int currentVersion = 1;
static AddressBook *singleInstance = nil;

NSString *PathToAddressBook()
{
  return [NSString stringWithFormat: @"%@/%@",
		   GNUMailUserLibraryPath(), @"AddressBook"];
}


//
//
//
@implementation AddressBook

- (id) init
{
  self = [super init];

  [AddressBook setVersion: currentVersion];

  [self setGroups: [NSArray array]];

  return self;
}


//
//
//
- (void) dealloc
{
  RELEASE(allGroups);

  [super dealloc];
}


//
//
//
- (BOOL) synchronize
{
  return [NSArchiver archiveRootObject: self toFile: PathToAddressBook()];
}


//
// NSCoding protocol
//
- (void) encodeWithCoder: (NSCoder *) theCoder
{
  [AddressBook setVersion: currentVersion];

  [theCoder encodeObject: [self allGroups]];
}

- (id) initWithCoder: (NSCoder *) theCoder
{
  int version;
  
  version = [theCoder versionForClassName: NSStringFromClass([self class])];;
  
  self = [super init];

  // Initial version of the serialized filter object
  if ( version == 0 )
    {
      NSArray *anArray;
      Group *aGroup;
      int i;

      aGroup = [[Group alloc] initWithName: _(@"General")];

      anArray = (NSArray *)[theCoder decodeObject];
      
      for (i = 0; i < [anArray count]; i++)
	{
	  [aGroup addAddress: [anArray objectAtIndex: i]];
	}
      
      [self setGroups: [NSArray arrayWithObject: aGroup]];
      RELEASE(aGroup);
    }
  // Version 1
  else
    {
      [self setGroups: [theCoder decodeObject]];
    }

  return self;
}


//
// access/mutation methods
//
- (void) addGroup: (Group *) theGroup
{
  [allGroups addObject: theGroup];
}


//
//
//
- (void) removeGroup: (Group *) theGroup;
{
  [allGroups removeObject: theGroup];
}


//
//
//
- (NSArray *) allGroups
{
  return allGroups;
}


//
//
//
- (void) setGroups: (NSArray *) theGroups
{
  if ( theGroups )
    {
      NSMutableArray *aMutableArray;

      aMutableArray = [[NSMutableArray alloc] initWithArray: theGroups];
      RELEASE(allGroups);
      allGroups = aMutableArray;
    }
  else
    {
      RELEASE(allGroups);
      allGroups = nil;
    }
}


//
//
//
- (Group *) groupAtIndex: (int) theIndex
{
  return [[self allGroups] objectAtIndex: theIndex];
}


//
//
//
- (Group *) groupForName: (NSString *) theName
{
  int i;

  for (i = 0; i < [allGroups count]; i++)
    {
      if ( [[(Group *)[allGroups objectAtIndex: i] name] caseInsensitiveCompare: theName] == NSOrderedSame)
	{
	  return [allGroups objectAtIndex: i];
	}
    }
  
  return nil;
}


//
//
//
- (unsigned int) indexOfGroup: (Group *) theGroup
{
  return [[self allGroups] indexOfObject: theGroup];
}


//
//
//
- (void) sort
{
  [allGroups sortUsingSelector: @selector(compareAccordingToName:)];
}


//
//
//
- (NSArray *) addressesWithPrefix: (NSString *) thePrefix
{
  NSMutableArray *aMutableArray;
  int i;
  
  if ( !thePrefix || 
       [thePrefix length] == 0 )
    {
      return [NSArray array];
    }
  
  
#ifdef MACOSX  // Use Apple's addressbook if desired; default operation
  if ( ([[NSUserDefaults standardUserDefaults] stringForKey: @"ADDRESSBOOK_TYPE"] == nil) ||
       ([[NSUserDefaults standardUserDefaults] integerForKey: @"ADDRESSBOOK_TYPE"] == APPLE_ADDRESSBOOK_TYPE) )
    {
      ABSearchElement *searchFirstName, *searchLastName, *searchEmail, *searchGroup, *compoundSearch;
      ABAddressBook *AB = [ABAddressBook sharedAddressBook];
      ABRecord *thisABRecord;

      NSEnumerator *abRecordEnumerator;
      NSMutableArray *hitArray;
      
      // Create our search elements; we search on first name, last name, and email address.
      // Do search on last name last so that the order of the list is sensible.
      searchFirstName = [ABPerson searchElementForProperty: kABFirstNameProperty
				  label: nil
				  key: nil
				  value: thePrefix
				  comparison: kABPrefixMatchCaseInsensitive];
      
      searchLastName = [ABPerson searchElementForProperty: kABLastNameProperty
				 label: nil
				 key: nil
				 value: thePrefix
				 comparison: kABPrefixMatchCaseInsensitive];
      
      searchEmail = [ABPerson searchElementForProperty: kABEmailProperty
			      label: nil
			      key: nil
			      value: thePrefix
			      comparison: kABPrefixMatchCaseInsensitive];
      
      // We also search on groups    
      searchGroup = [ABGroup searchElementForProperty: kABGroupNameProperty
			     label: nil
			     key:nil
			     value: thePrefix
			     comparison: kABPrefixMatchCaseInsensitive];
      
      
      
      compoundSearch = [ABSearchElement searchElementForConjunction: kABSearchOr
					children: [NSArray arrayWithObjects: searchFirstName, 
							   searchEmail, searchGroup, nil]];
      
      hitArray = [[NSMutableArray alloc] init];
      AUTORELEASE(hitArray);
      aMutableArray = [[NSMutableArray alloc] init];
      
      [hitArray addObjectsFromArray: [AB recordsMatchingSearchElement: compoundSearch]];
      
      // Now add the results from search on last name.
      [hitArray addObjectsFromArray: [AB recordsMatchingSearchElement: searchLastName]];
      abRecordEnumerator = [hitArray objectEnumerator];
      
      while ( (thisABRecord = [abRecordEnumerator nextObject]) )
        {        
	  // check if this is a group record
	  if ( [thisABRecord isKindOfClass: [ABGroup class]] ) 
            {
	      Group *aGroup;
	      
	      aGroup = [[Group alloc] init];
	      
	      [aGroup setName: [NSString stringWithString: [thisABRecord valueForProperty: kABGroupNameProperty]]];
	      [aMutableArray addObject: aGroup];  
	      RELEASE(aGroup);   
            }
	  // this is a person record
	  else 
            {
	      Address *anAddress;
	      int emailCount;
	      
	      emailCount = [[thisABRecord valueForProperty: kABEmailProperty] count];
	      
	      for (i = 0; i < emailCount; i++)
                {
		  anAddress = [[Address alloc] init];
		  
                  if ( [thisABRecord valueForProperty: kABLastNameProperty] )
                    {
		      [anAddress setName: [NSString stringWithFormat:@"%@ %@", 
						    [thisABRecord valueForProperty: kABFirstNameProperty],
						    [thisABRecord valueForProperty: kABLastNameProperty]]];
                    }
                  else
                    {
                      [anAddress setName: [thisABRecord valueForProperty: kABFirstNameProperty]];
                    }
		  
		  [anAddress setEmailAddress: [[thisABRecord valueForProperty: kABEmailProperty] valueAtIndex: i]];
		  [aMutableArray addObject: anAddress];  
		  RELEASE(anAddress);
                }        
            }
	  
        }
      
      return AUTORELEASE(aMutableArray);    
    }      
#endif
  
  // GNUMail.app Addressbook
  aMutableArray = [[NSMutableArray alloc] init];
  
  for (i = 0; i < [allGroups count]; i++)
    {
      Group *aGroup;
      int j;
      
      aGroup = [self groupAtIndex: i];
      
       if ( [[[aGroup name] lowercaseString] hasPrefix: [thePrefix lowercaseString]] )
      	{
      	  [aMutableArray addObject: aGroup];
      	}
      
      for (j = 0; j < [[aGroup addresses] count]; j++)
	{
	  Address *anAddress;
	  
	  anAddress = [aGroup addressAtIndex: j];
	  
	  if ( [anAddress lowercaseName] && [[anAddress lowercaseName] length] > 0 &&
	       [[anAddress lowercaseName] hasPrefix: [thePrefix lowercaseString]] )
	    {
	      [aMutableArray addObject: anAddress];
	    }
	  //else if ( (![anAddress lowercaseName] || ([anAddress lowercaseName] && [[anAddress lowercaseName] length] == 0)) &&
	  //	    [anAddress emailAddress] && [[anAddress emailAddress] length] > 0 && 
	  //	    [[[anAddress emailAddress] lowercaseString] hasPrefix: [thePrefix lowercaseString]] )
	  // {
	  //  [aMutableArray addObject: anAddress];
	  //}
	}
    }
 
  return AUTORELEASE(aMutableArray);
}


//
//
//
- (NSArray *) addressesWithSubstring: (NSString *) theSubstring
{
  NSMutableArray *aMutableArray;
  int i;
  
  if ( !theSubstring || 
       [theSubstring length] == 0 )
    {
      return [NSArray array];
    }
  
  aMutableArray = [[NSMutableArray alloc] init];
  
  for (i = 0; i < [allGroups count]; i++)
    {
      Group *aGroup;
      int j;

      aGroup = [self groupAtIndex: i];

      for (j = 0; j < [[aGroup addresses] count]; j++)
	{
	  Address *anAddress;

	  anAddress = [aGroup addressAtIndex: j];

	  if ( [anAddress lowercaseName] && [[anAddress lowercaseName] length] > 0 )
	    {
	      NSRange aRange;

	      aRange = [[theSubstring lowercaseString] rangeOfString: [anAddress lowercaseName]];

	      if (aRange.length)
		{
		  [aMutableArray addObject: anAddress];
		  continue;
		}
	    }
	  else if ( (![anAddress lowercaseName] || ([anAddress lowercaseName] && [[anAddress lowercaseName] length] == 0)) &&
		    [anAddress emailAddress] )
	    {
	      NSRange aRange;

	      aRange = [[theSubstring lowercaseString] rangeOfString: [anAddress emailAddress]];

	      if (aRange.length)
		{
		  [aMutableArray addObject: anAddress];
		  continue;
		}
	    }
	}
    }

  return AUTORELEASE(aMutableArray);
}



//
// This method verifies that the archive is valid so we won't
// segfault in case the archive is corrupted.
//
+ (id) singleInstance
{
  if ( singleInstance == nil )
    {      
      NS_DURING
	singleInstance = [NSUnarchiver unarchiveObjectWithFile: PathToAddressBook()];
      NS_HANDLER
	NSLog(@"Caught exception while unarchiving the AddressBook. Ignoring.");
        singleInstance = nil;
      NS_ENDHANDLER
	
      if ( singleInstance )
	{
	  RETAIN(singleInstance);
	}
      else
	{
	  singleInstance = [[AddressBook alloc] init];
	  [singleInstance synchronize];
	}
    }
  
  return singleInstance;
}

@end
