/*
**  AddressBookController.m
**
**  Copyright (c) 2001, 2002, 2003
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**          Jonathan B. Leffert <jonathan@leffert.net>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AddressBookController.h"

#ifdef MACOSX
#import <AddressBook/AddressBook.h>
#endif

#include "Address.h"
#include "AddressBook.h"
#include "AddressTaker.h"

#ifndef MACOSX
#include "AddressBookWindow.h"
#endif

#include "GNUMail.h"
#include "Constants.h"
#include "Group.h"
#include "Utilities.h"

#include <Pantomime/InternetAddress.h>
#include <Pantomime/NSString+Extensions.h>

static AddressBookController *singleInstance = nil;

@implementation AddressBookController

// we use the window ivar to represent the address book window, but we
// expect only a single instance of the controller app-wide.

- (id) initWithWindowNibName: (NSString *) windowNibName
{

#ifdef MACOSX
  
  self = [super initWithWindowNibName: windowNibName];
 
#else
  AddressBookWindow *theWindow;
 
  theWindow =
    [[AddressBookWindow alloc] initWithContentRect: NSMakeRect(200,200,480,325)
			       styleMask: (NSTitledWindowMask |
					   NSClosableWindowMask |
					   NSMiniaturizableWindowMask)
			       backing: NSBackingStoreBuffered
			       defer: YES];

  self = [super initWithWindow: theWindow];

  [theWindow layoutWindow];
  [theWindow setDelegate: self];
  [theWindow setMinSize: [theWindow frame].size];

  // We link our outlets
  browser = theWindow->browser;
  nameTextField = theWindow->nameTextField;
  add = theWindow->add;
  delete = theWindow->delete;
  modify = theWindow->modify;
  moveToPopUp = theWindow->moveToPopUp;
  importPopUp = theWindow->importPopUp;

  RELEASE(theWindow);
#endif

  // We set some initial parameters
  [[self window] setTitle: _(@"Address Book")];

  // Initially, we disable most buttons
  [add setEnabled: YES];
  [add setAction: @selector(addGroup:)];
  [add setTitle: _(@"Add Group")];
  
  [delete setEnabled: NO];
  [delete setTitle: _(@"Delete Group")];
  
  [modify setEnabled: NO];
  
  [moveToPopUp setEnabled: NO];

#ifdef MACOSX
  [browser setAction: @selector(selectionInBrowserHasChanged:)];
  [browser setDoubleAction: @selector(doubleClicked:)];
#endif

  [self _loadItemsInMoveToPopUpButton];

  // We finally set our autosave window frame name and restore the one from the user's defaults.
  [[self window] setFrameAutosaveName: @"AddressBookWindow"];
  [[self window] setFrameUsingName: @"AddressBookWindow"];
  
  return self;
}


//
// This method is used to add an address to the address book
// from an 'external' object. It searches for the group
// "General" (localized) and if it can't find it, it creates a new one.
//
- (void) addToAddressBook: (Address *) theAddress
{
  if ( theAddress &&
       [theAddress isKindOfClass: [Address class]] )
    {
      NSArray *allGroups;
      Group *aGroup;
      int i;
#ifdef MACOSX  // Use Apple's addressbook if desired; default operation
      if( ([[NSUserDefaults standardUserDefaults] stringForKey: @"ADDRESSBOOK_TYPE"] == nil) ||
            ([[NSUserDefaults standardUserDefaults] integerForKey: @"ADDRESSBOOK_TYPE"] == APPLE_ADDRESSBOOK_TYPE) )
        {
	  ABAddressBook *AB;
	  ABPerson *aPerson;
	  NSString *firstName, *lastName;
	  ABMutableMultiValue *anAddressList;
	  NSRange aRange;
	  
	  // Allocate a new record
	  aPerson = [[ABPerson alloc] init];
	  
	  // Parse out the first and last names
	  if ( [[theAddress name] length] > 0 )
            {
	      aRange = [[theAddress name] rangeOfString: @" " options: NSBackwardsSearch];
	      if ( aRange.length > 0 )
                {
		  firstName = [[theAddress name] substringToIndex: aRange.location];
		  lastName = [[theAddress name] substringFromIndex: aRange.location+1];
		  [aPerson setValue: firstName forProperty: kABFirstNameProperty];
		  [aPerson setValue: lastName forProperty: kABLastNameProperty];
                }
	      else
                {
		  [aPerson setValue: [theAddress name] forProperty: kABFirstNameProperty];
                }
            }
	  
	  // Set the email address
	  if ( [[theAddress emailAddress] length] > 0 )
            {
	      anAddressList = [[ABMutableMultiValue alloc] init];
	      [anAddressList addValue: [theAddress emailAddress] withLabel: kABEmailWorkLabel];
	      [aPerson setValue: anAddressList forProperty: kABEmailProperty];
	      RELEASE(anAddressList);
            }
	  
	  // Add the record
	  AB = [ABAddressBook sharedAddressBook];
	  [AB addRecord: aPerson];
	  [AB save];
	  RELEASE(aPerson);
	  
	  return;
        }
#endif      
      
      allGroups = [[AddressBook singleInstance] allGroups];
      aGroup = nil;
      
      for (i = 0; i < [allGroups count]; i++)
	{
	  aGroup = [allGroups objectAtIndex: i];
	  
	  if ([[aGroup name] caseInsensitiveCompare: _(@"General")] == NSOrderedSame)
	    {
	      break;
	    }

	  aGroup = nil;
	}

      // We haven't found a group "General", let's create one
      if ( !aGroup )
	{
	  aGroup = [[Group alloc] initWithName: _(@"General")];
	  [[AddressBook singleInstance] addGroup: aGroup];
          [[AddressBook singleInstance] sort];
	}

      [aGroup addAddress: theAddress];
      [aGroup sort];

      [[AddressBook singleInstance] synchronize];

      // Let's refresh our browser
      [browser loadColumnZero];
    }
  else
    {
      NSBeep();
    }
}

//
// action methods
//
- (IBAction) addAddress: (id) sender
{
  NSString *aString;
  
  aString = [[nameTextField stringValue] stringByTrimmingWhiteSpaces];

  if ( [aString length] )
    {
      InternetAddress *anInternetAddress;
      Address *anAddress;
      NSArray *allGroups;
      Group *aGroup;

      allGroups = [[AddressBook singleInstance] allGroups];
      
      aGroup = [allGroups objectAtIndex: [browser selectedRowInColumn: 0]];
      
      anInternetAddress = [[InternetAddress alloc] initWithString: aString];
      anAddress = [[Address alloc] init];

      [anAddress setName: [anInternetAddress personal]];
      [anAddress setEmailAddress: [anInternetAddress address]];
      
      [aGroup addAddress: anAddress];

      RELEASE(anInternetAddress);
      RELEASE(anAddress);

      [aGroup sort];

      [[AddressBook singleInstance] synchronize];
     
      [browser reloadColumn: 1];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) addGroup: (id) sender
{
  NSString *aString;
  
  aString = [[nameTextField stringValue] stringByTrimmingWhiteSpaces];

  if ( [aString length] )
    {
      NSArray *allGroups;
      Group *aGroup;
      int i;

      allGroups = [[AddressBook singleInstance] allGroups];
      
      // We verify that the group isn't already there.
      for (i = 0; i < [allGroups count]; i++)
	{
	  aGroup = [allGroups objectAtIndex: i];

	  if ( [[aGroup name] caseInsensitiveCompare: aString] == NSOrderedSame )
	    {
	      NSRunInformationalAlertPanel(_(@"Error!"),
					   _(@"This group already exist! Please choose an other name."),
					   _(@"OK"),
					   NULL, 
					   NULL,
					   NULL);
	      return;
	    }
	}

      // Everything is alright, let's add our new group
      aGroup = [[Group alloc] initWithName: aString];
      
      [[AddressBook singleInstance] addGroup: aGroup];
      [[AddressBook singleInstance] sort];
      [[AddressBook singleInstance] synchronize];
      
      RELEASE(aGroup);

      [browser loadColumnZero];
      [self _loadItemsInMoveToPopUpButton];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) deleteAddress: (id) sender
{
  if ( [browser selectedCellInColumn: 1] )
    {
      int choice;
      
      choice = NSRunAlertPanel(_(@"Delete..."),
			       _(@"Are you sure you want to delete this address?"),
			       _(@"Cancel"),     // default
			       _(@"Delete"),    // alternate
			       NULL);        // other button
      
      
      if (choice == NSAlertAlternateReturn)
	{
	  Group *aGroup;
	  
	  aGroup = [[[AddressBook singleInstance] allGroups] objectAtIndex: [browser selectedRowInColumn: 0]];
	  
	  // We verify to be sure we aren't trying to delete our fake address
	  if ( [[aGroup addresses] count] == 0)
	    {
	      NSBeep();
	    }
	  else
	    {
	      Address *anAddress;
	      
	      anAddress = [[aGroup addresses] objectAtIndex: [browser selectedRowInColumn: 1]];

	      [aGroup removeAddress: anAddress];
	      [[AddressBook singleInstance] synchronize];
	      
	      [browser reloadColumn: 1];
	    }
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) deleteGroup: (id) sender
{
  
  if ( [browser selectedCellInColumn: 0] )
    {
      int choice;
      
      choice = NSRunAlertPanel(_(@"Delete..."),
			       _(@"Are you sure you want to delete this group?"),
			       _(@"Cancel"),     // default
			       _(@"Delete"),    // alternate
			       NULL);        // other button
      
      
      if (choice == NSAlertAlternateReturn)
	{
	  Group *aGroup;
	  
	  aGroup = [[[AddressBook singleInstance] allGroups] objectAtIndex: [browser selectedRowInColumn: 0]];
	  
	  [[AddressBook singleInstance] removeGroup: aGroup];
	  [[AddressBook singleInstance] synchronize];
	  
	  [browser loadColumnZero];
	  [self _loadItemsInMoveToPopUpButton];
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) doubleClicked: (id) sender
{
  int modifier = [[[sender window] currentEvent] modifierFlags];

  if ( (modifier & NSControlKeyMask) &&
       ! (modifier & NSShiftKeyMask) )
    {
      return [self ccClicked: sender];
    }
  else if ( ! (modifier & NSControlKeyMask) &&
            (modifier & NSShiftKeyMask) )
    {
      return [self bccClicked: sender];
    }
  else
    {
      return [self toClicked: sender];
    }
}


//
//
//
- (IBAction) toClicked: (id) sender
{
  // We selected an address...
  if ([browser selectedCellInColumn: 1] &&
      [GNUMail lastAddressTakerWindowOnTop] )
    {
      unsigned int row;
      
      row = [browser selectedRowInColumn: 1];

      if ( row >= 0 )
	{
	  NSArray *allGroups, *allAddresses;
	  
	  allGroups = [[AddressBook singleInstance] allGroups];
	  allAddresses = [[allGroups objectAtIndex: [browser selectedRowInColumn: 0]] addresses];

	  if ( [allAddresses count] > 0 )
	    {
	      [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
				 takeToAddress: [allAddresses objectAtIndex: row]];
	    }
	  else
	    {
	      NSBeep();
	    }
	}
      else
	{
	  NSBeep();
	}
    }
  // We selected a group
  else if ([browser selectedCellInColumn: 0] &&
	   [GNUMail lastAddressTakerWindowOnTop] )
    {
      NSArray *allAddresses;
      Group *aGroup;
      int i;

      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
      allAddresses = [aGroup addresses];

      for (i = 0; i < [allAddresses count]; i++)
	{
	  [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
			     takeToAddress: [allAddresses objectAtIndex: i]];
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) ccClicked: (id) sender
{
  // We selected an address...
  if ([browser selectedCellInColumn: 1] &&
      [GNUMail lastAddressTakerWindowOnTop] )
    {
      unsigned int row;
      
      row = [browser selectedRowInColumn: 1];

      if ( row >= 0 )
	{
	  NSArray *allGroups, *allAddresses;
	  
	  allGroups = [[AddressBook singleInstance] allGroups];
	  allAddresses = [[allGroups objectAtIndex: [browser selectedRowInColumn: 0]] addresses];

	  if ( [allAddresses count] > 0 )
	    {
	      [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
				 takeCcAddress: [allAddresses objectAtIndex: row]];
	    }
	  else
	    {
	      NSBeep();
	    }
	}
      else
	{
	  NSBeep();
	}
    }
  // We selected a group
  else if ([browser selectedCellInColumn: 0] &&
	   [GNUMail lastAddressTakerWindowOnTop] )
    {
      NSArray *allAddresses;
      Group *aGroup;
      int i;

      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
      allAddresses = [aGroup addresses];

      for (i = 0; i < [allAddresses count]; i++)
	{
	  [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
			     takeCcAddress: [allAddresses objectAtIndex: i]];
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) bccClicked: (id) sender
{
  // We selected an address...
  if ([browser selectedCellInColumn: 1] &&
      [GNUMail lastAddressTakerWindowOnTop] )
    {
      unsigned int row;
      
      row = [browser selectedRowInColumn: 1];

      if ( row >= 0 )
	{
	  NSArray *allGroups, *allAddresses;
	  
	  allGroups = [[AddressBook singleInstance] allGroups];
	  allAddresses = [[allGroups objectAtIndex: [browser selectedRowInColumn: 0]] addresses];

	  if ( [allAddresses count] > 0 )
	    {
	      [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
				 takeBccAddress: [allAddresses objectAtIndex: row]];
	    }
	  else
	    {
	      NSBeep();
	    }
	}
      else
	{
	  NSBeep();
	}
    }
  // We selected a group
  else if ([browser selectedCellInColumn: 0] &&
	   [GNUMail lastAddressTakerWindowOnTop] )
    {
      NSArray *allAddresses;
      Group *aGroup;
      int i;

      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
      allAddresses = [aGroup addresses];

      for (i = 0; i < [allAddresses count]; i++)
	{
	  [(id<AddressTaker>)[GNUMail lastAddressTakerWindowOnTop]
			     takeBccAddress: [allAddresses objectAtIndex: i]];
	}
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) modifyAddress: (id) sender
{
  NSString *aString;
  
  aString = [[nameTextField stringValue] stringByTrimmingWhiteSpaces];

  if ( [aString length] &&
       [browser selectedCellInColumn: 1] )
    {
      Group *aGroup;
      
      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
      
      if ( [[aGroup addresses] count] > 0 )
	{
	  InternetAddress *anInternetAddress;
	  Address *anAddress;

	  anAddress = [aGroup addressAtIndex: [browser selectedRowInColumn: 1]];
	  
	  anInternetAddress = [[InternetAddress alloc] initWithString: aString];
	  [anAddress setName: [anInternetAddress personal]];
	  [anAddress setEmailAddress: [anInternetAddress address]];
	  
	  RELEASE(anInternetAddress);
	  
	  [aGroup sort];

	  [[AddressBook singleInstance] synchronize];
	  
	  [browser reloadColumn: 1];
	}
      else
	{
	  NSBeep();
	}
    }
  else
    {
      NSBeep();
    }

}


//
//
//
- (IBAction) modifyGroup: (id) sender
{
  NSString *aString;

  aString = [[nameTextField stringValue] stringByTrimmingWhiteSpaces];

  if ( [aString length] > 0 &&
       [browser selectedCellInColumn: 0] )
    {
      Group *aGroup;
      
      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];

      [aGroup setName: aString];
      [[AddressBook singleInstance] sort];
      [[AddressBook singleInstance] synchronize];

      [browser loadColumnZero];
      [self _loadItemsInMoveToPopUpButton];
    }
  else
    {
      NSBeep();
    }
}


//
//
//
- (IBAction) selectionInImportPopUpHasChanged: (id) sender
{
  int index;

  [importPopUp synchronizeTitleAndSelectedItem];

  // index == 0 -> "Import from:"
  // index == 1 -> "OPENSTEP Mail.app"
  // index == 2 -> "LDIF"
  index = [importPopUp indexOfSelectedItem];

  if ( index <= 0 )
    {
      NSBeep();
      return;
    }      
  
  // OPENSTEP Mail.app
  if ( index == 1 )
    {
      [self _importFromOPENSTEPMail];
    }
  // LDIF
  else if ( index == 2)
    {
      Group *aGroup;
      
      // We first verify if we selected a group
      if ( ![browser selectedCellInColumn: 0] )
	{
	  NSRunAlertPanel(_(@"Error!"),
			  _(@"You must select a group to import the addresses in."),
			  _(@"OK"), // default
			  NULL,     // alternate
			  NULL);
	  return;
	}
      
      aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];

      [self _importFromLDIFInGroup: aGroup];
    }
}


//
//
//
- (IBAction) selectionInMoveToPopUpHasChanged: (id) sender
{
  [moveToPopUp synchronizeTitleAndSelectedItem];
  
  if ( [moveToPopUp indexOfSelectedItem] <= 0 ||
       ![browser selectedCellInColumn: 1] )
    {
      NSBeep();
    }
  else
    {
      Group *selectedGroup, *aGroup;
      Address *anAddress;

      selectedGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
      aGroup = [[AddressBook singleInstance] groupForName: [moveToPopUp titleOfSelectedItem]];

      // We verify that we aren't trying to transfer to the same group and that
      // we aren't trying to transfer our fake address
      if (aGroup == selectedGroup ||
	  [[selectedGroup addresses] count] == 0)
	{
	  NSBeep();
	  return;
	}
      
      // Everything is alright, let's move the address!
      anAddress = [selectedGroup addressAtIndex: [browser selectedRowInColumn: 1]];

      [aGroup addAddress: anAddress];
      [selectedGroup removeAddress: anAddress];
      
      [[AddressBook singleInstance] synchronize];

      [browser reloadColumn: 1];
      [nameTextField setStringValue: @""];

      
      [moveToPopUp selectItemAtIndex: 0];
      [moveToPopUp synchronizeTitleAndSelectedItem];
    }
}


//
//
//
- (IBAction) selectionInBrowserHasChanged: (id) sender
{
  int row, column;

  if ( [browser selectedCellInColumn: 1] )
    {
      row = [browser selectedRowInColumn: 1];
      column = 1;
    }
  else
    {
      row = [browser selectedRowInColumn: 0];
      column = 0;
    }
  
  [self browser: browser
	selectRow: row
	inColumn: column];
}



//
// delegate methods
//
- (void) windowDidLoad
{
#ifdef MACOSX   
  [browser loadColumnZero];
  [browser setNeedsDisplay: YES];
#endif
}


//
//
//
- (int)       browser: (NSBrowser *) sender
 numberOfRowsInColumn: (int) column
{
  if (column == 0)
    {
      return [[[AddressBook singleInstance] allGroups] count];
    }
  else if (column == 1)
    {
      NSArray *allGroups;
      Group *aGroup;
      int count;

      allGroups = [[AddressBook singleInstance] allGroups];
      aGroup = [allGroups objectAtIndex: [browser selectedRowInColumn: 0]];

      count = [[aGroup addresses] count];
 
      // We ALWAYS return at least one 'fake address'
      return (count == 0 ? 1 : count);
    }
  
  return 0;
}


//
//
//
- (void) browser: (NSBrowser *) sender
 willDisplayCell: (id) cell
	   atRow: (int) row
	  column: (int) column
{
  if (column == 0)
    {
      NSArray *allGroups;
      Group *aGroup;
      
      allGroups = [[AddressBook singleInstance] allGroups];
      aGroup = [allGroups objectAtIndex: row];
      
      [cell setStringValue: [aGroup name]];
    }
  else if (column == 1)
    {
      NSArray *allGroups, *allAddresses;

      allGroups = [[AddressBook singleInstance] allGroups];   
      allAddresses = [[allGroups objectAtIndex: [browser selectedRowInColumn: 0]] addresses];
      
      // We verify if we need to create a 'fake address'
      if ( [allAddresses count] == 0 )
	{
	  [cell setStringValue: _(@"< no address >")];
	}
      else
	{
	  [cell setStringValue: [[allAddresses objectAtIndex: row] description]];
	}
      
      [cell setLeaf: YES];
    }
}

- (BOOL) browser: (NSBrowser *) sender
       selectRow: (int) row
        inColumn: (int) column
{
  [add setEnabled: YES];
  [delete setEnabled: YES];
  [modify setEnabled: YES];

  if (column == 0)
    {
      NSArray *allGroups;
      Group *aGroup;

      allGroups = [[AddressBook singleInstance] allGroups];
      
      aGroup = [allGroups objectAtIndex: row];
      [aGroup sort];

      [moveToPopUp setEnabled: NO];

      [add setTitle: _(@"Add Group")];
      [add setAction: @selector(addGroup:)];

      [delete setTitle: _(@"Delete Group")];
      [delete setAction: @selector(deleteGroup:)];

      [modify setAction: @selector(modifyGroup:)];
      
      [nameTextField setStringValue: [aGroup name]];
    }
  else if (column == 1)
    {
      NSArray *allGroups, *allAddresses;

      allGroups = [[AddressBook singleInstance] allGroups];
      allAddresses = [[allGroups objectAtIndex: [browser selectedRowInColumn: 0]] addresses];

      [moveToPopUp setEnabled: YES];

      [add setTitle: _(@"Add Address")];
      [add setAction: @selector(addAddress:)];

      [delete setTitle: _(@"Delete Address")];
      [delete setAction: @selector(deleteAddress:)];

      [modify setAction: @selector(modifyAddress:)];

      // We verify if we selected our fake addresse
      if ( [allAddresses count] == 0 )
	{
	  [nameTextField setStringValue: _(@"Enter the contact's E-Mail address here")];
	}
      else
	{
	  [nameTextField setStringValue: [[allAddresses objectAtIndex: row] description]];
	}
    }

  return YES;
}

- (NSString *) browser: (NSBrowser *) sender
         titleOfColumn: (int) column
{
  if (column == 0)
    {
      return _(@"Groups");
    }
  else
    {
      if ( [browser selectedCellInColumn: 0] )
	{
	  Group *aGroup;
	  
	  aGroup = [[AddressBook singleInstance] groupAtIndex: [browser selectedRowInColumn: 0]];
	 
	  return [aGroup name];
	}
    }

  return @"";
}


//
// access/mutation methods
//
- (NSBrowser *) browser
{
  return browser;
}


//
// class methods
//
+ (id) singleInstance
{
  if ( !singleInstance )
    {
      singleInstance = [[AddressBookController alloc] initWithWindowNibName: @"AddressBookWindow"];
    }
  
  return singleInstance;
}

@end


//
// Private methods
//
@implementation AddressBookController (Private)

- (void) _importFromOPENSTEPMail
{
  NSOpenPanel *oPanel;
  int result;
  
  oPanel = [NSOpenPanel openPanel];
  [oPanel setAllowsMultipleSelection:NO];
  result = [oPanel runModalForDirectory: NSHomeDirectory()
		   file: nil 
		   types: nil];
  
  if (result == NSOKButton) 
    {
      NSArray *fileToOpen;
      
      fileToOpen = [oPanel filenames];
      
      if ([fileToOpen count] > 0)
	{
	  NSArray *allLines;
	  NSString *aString;
	  int i;

	  aString = [NSString stringWithContentsOfFile: [fileToOpen objectAtIndex: 0]];
	  
	  allLines = [aString componentsSeparatedByString: @"\n"];

	  for (i = 0; i < [allLines count]; i++)
	    {	      
	      NSString *aLine;
	      NSRange aRange;
 
	      aLine = [allLines objectAtIndex: i];	      
	      aRange = [aLine rangeOfString: @":"];
	      
	      // We verify if we've found a good line.
	      if ( aRange.length )
		{
		  NSArray *allAddresses;
		  NSString *aGroupName;
		  int j;
		  
		  Group *aGroup;

		  aGroupName = [aLine substringWithRange: NSMakeRange(0, aRange.location)];

		  // We got the group name, let's see if it already exists in our address book.
		  // We add it if it doesn't.
		  aGroup = [[AddressBook singleInstance] groupForName: aGroupName];
		  
		  if ( !aGroup )
		    {
		      aGroup = [[Group alloc] initWithName: aGroupName];
		      [[AddressBook singleInstance] addGroup: aGroup];
		      AUTORELEASE(aGroup);
		    }

		  allAddresses = [[aLine substringFromIndex: aRange.location + 1]
				   componentsSeparatedByString: @","];
		  
		  for (j = 0; j < [allAddresses count]; j++)
		    {
		      InternetAddress *anInternetAddress;
		      Address *anAddress;
		      
		      anInternetAddress = [[InternetAddress alloc] initWithString: 
								     [allAddresses objectAtIndex: j]];

		      anAddress = [[Address alloc] init];
		      [anAddress setName: [anInternetAddress personal]];
		      [anAddress setEmailAddress: [anInternetAddress address]];
		      
		      [aGroup addAddress: anAddress];

		      RELEASE(anInternetAddress);
		      RELEASE(anAddress);
		    }

		  [browser loadColumnZero];
		  [[AddressBook singleInstance] synchronize];
		}
	      else
		{
		  NSDebugLog(@"Improperly formatted line. Ignored.");
		}
	    }
	}
    }
}


//
//
//
- (void) _importFromLDIFInGroup: (Group *) theGroup
{
  NSOpenPanel *oPanel;
  int result;
  
  oPanel = [NSOpenPanel openPanel];
  [oPanel setAllowsMultipleSelection:NO];
  result = [oPanel runModalForDirectory: NSHomeDirectory()
		   file: nil 
		   types: nil];
  
  if (result == NSOKButton) 
    {
      NSArray *fileToOpen;
      
      fileToOpen = [oPanel filenames];
      
      if ([fileToOpen count] > 0)
	{
	  NSString *aString, *aName, *anAddress;
	  NSArray *allLines;
	  int i;

	  aString = [NSString stringWithContentsOfFile: [fileToOpen objectAtIndex: 0]];
	  allLines = [aString componentsSeparatedByString: @"\n"];
	  aName = nil;
	  anAddress = nil;

	  for (i = 0; i < [allLines count]; i++)
	    {	      
	      NSString *aLine;
	      NSRange aRange;
 
	      aLine = [allLines objectAtIndex: i];	      
	      aRange = [aLine rangeOfString: @":"];

	      // We verify if we've found a good line.
	      if ( aRange.length )
		{
		  if ( [aLine hasCaseInsensitivePrefix: @"cn:"] &&
		       [aLine length] > (aRange.location + 2) )
		    {
		      aName = [aLine substringFromIndex: aRange.location + 2];
		    }
		  else if ( [aLine hasCaseInsensitivePrefix: @"mail:"] &&
			    [aLine length] > (aRange.location + 2) )
		    {
		      anAddress = [aLine substringFromIndex: aRange.location + 2];
		    }
		}
	      else
		{
		  if (aName || anAddress )
		    {
		      Address *theAddress;
		      
		      theAddress = [[Address alloc] init];
		      [theAddress setName: (aName != nil ? aName : @"")];
		      [theAddress setEmailAddress: (theAddress != nil ? anAddress : @"")];
		      
		      [theGroup addAddress: theAddress];
		      RELEASE(theAddress);

		      [[AddressBook singleInstance] synchronize];
		    }

		  aName = nil;
		  anAddress = nil;
		}
	    }
	  
	  // We reload the content of our group column (the addresses)
	  [browser reloadColumn: 1];
	}
    }
}


//
//
//
- (void) _loadItemsInMoveToPopUpButton
{
  NSArray *allGroups;
  int i;

  allGroups = [[AddressBook singleInstance] allGroups];

  [moveToPopUp removeAllItems];
  [moveToPopUp addItemWithTitle: _(@"Move to:")];

  for (i = 0; i < [allGroups count]; i++)
    {
      Group *aGroup;

      aGroup = [allGroups objectAtIndex: i];
      
      [moveToPopUp addItemWithTitle: [aGroup name]];
    }
}

@end
