/*
**  AdvancedViewController.m
**
**  Copyright (c) 2002, 2003
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>   
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AdvancedViewController.h"

#include "Constants.h"
#include "ExtendedTableColumn.h"
#include "SwitchTableView.h"

#ifndef MACOSX
#include "AdvancedView.h"
#endif


static AdvancedViewController *singleInstance = nil;

static struct { NSString *option; NSString *key; int value; } rows[] = {
  {@"Compact mailbox when closing", @"COMPACT_MAILBOX_ON_CLOSE", NSOffState},
  {@"Prompt before compacting a mailbox", @"PROMPT_BEFORE_COMPACT", NSOffState},
  {@"Repeat the message's subject (message threading)", @"MAILWINDOW_REPEAT_SUBJECT", NSOffState},
  {@"Automatically expand threads (message threading)", @"AUTOMATICALLY_EXPAND_THREADS", NSOffState},
  {@"Open the Console window on startup", @"OPEN_CONSOLE_ON_STARTUP", NSOffState},
  {@"Open the Mailboxes window on startup", @"OPEN_MAILBOXMANAGER_ON_STARTUP", NSOffState},
  {@"Do not render messages in the MailWindow", @"NO_MESSAGE_RENDER_IN_MAIL_WINDOW", NSOffState}
};


//
//
//
@implementation AdvancedViewController

- (id) initWithNibName: (NSString *) theName
{
  self = [super init];

#ifdef MACOSX
  
  if ( ![NSBundle loadNibNamed: theName
		  owner: self] )
    {
      NSDebugLog(@"Fatal error occured while loading the AdvancedView nib file");
      AUTORELEASE(self);
      return nil;
    }

  RETAIN(view);

#else
  // We link our view
  view = [[AdvancedView alloc] initWithParent: self];
  [view layoutView];

  // We link our outlets
  tableView = ((AdvancedView*)view)->tableView;
  optionsColumn = ((AdvancedView*)view)->optionsColumn;
  enabledColumn = ((AdvancedView*)view)->enabledColumn;
#endif

#ifdef MACOSX
  {
    ExtendedTableColumn *extendedTableColumn;
    NSButtonCell *cell;

    cell = [[NSButtonCell alloc] init];

    [cell setButtonType: NSSwitchButton];
    [cell setImagePosition: NSImageOnly];
    [cell setControlSize: NSSmallControlSize];
    
    extendedTableColumn = (ExtendedTableColumn *)[tableView tableColumnWithIdentifier: @"enabled"];
    [extendedTableColumn setDataCell: cell];
    [extendedTableColumn setShouldUseMouse: YES];
    [extendedTableColumn setShouldUseAndSetState: YES];
    
    RELEASE(cell);
  }
#endif

  // We get our defaults for this panel
  [self initializeFromDefaults];

  return self;
}


//
//
//
- (void) dealloc
{
  NSDebugLog(@"AdvancedViewController: -dealloc");

  // Cocoa bug?
#ifdef MACOSX
  [tableView setDataSource: nil];
#endif

  singleInstance = nil;
  RELEASE(view);

  [super dealloc];
}


//
// access methods
//
- (NSImage *) image
{
  NSBundle *aBundle;
  
  aBundle = [NSBundle bundleForClass: [self class]];
  
  return AUTORELEASE([[NSImage alloc] initWithContentsOfFile:
					[aBundle pathForResource: @"advanced" ofType: @"tiff"]]);
}


//
//
//
- (NSString *) name
{
  return _(@"Advanced");
}


//
//
//
- (NSView *) view
{
  return view;
}

- (BOOL) hasChangesPending
{
  return YES;
}


//
//
//
- (void) initializeFromDefaults
{
  int i;

  for (i = 0; i < sizeof(rows)/sizeof(rows[0]); i++)
    {
      if ( [[NSUserDefaults standardUserDefaults] objectForKey: rows[i].key] )
	{
	  rows[i].value = [[NSUserDefaults standardUserDefaults] integerForKey: rows[i].key];
	}
    }
}


//
//
//
- (void) saveChanges
{ 
  int i;

  for (i = 0; i < sizeof(rows)/sizeof(rows[0]); i++)
    {
      [[NSUserDefaults standardUserDefaults] setInteger: rows[i].value
					     forKey: rows[i].key];
    }
}


//
// Data Source methods
//
- (int) numberOfRowsInTableView: (NSTableView *)aTableView
{
  return sizeof(rows)/sizeof(rows[0]);
}


//
//
//
- (id)           tableView: (NSTableView *)aTableView
 objectValueForTableColumn: (NSTableColumn *)aTableColumn 
		       row: (int)rowIndex
{
  if ( aTableColumn == optionsColumn )
    {
      return rows[rowIndex].option;
    }

  return ( rows[rowIndex].value == NSOnState ? _(@"Yes") : _(@"No") );
}


- (int)    tableView: (NSTableView *) aTableView
 stateForTableColumn: (NSTableColumn *) aTableColumn 
		 row: (int) rowIndex
{
  return rows[rowIndex].value;
}


- (void) tableView: (NSTableView *) aTableView
	  setState: (int) aState
    forTableColumn: (NSTableColumn *) aTableColumn 
	       row: (int) rowIndex
{
  rows[rowIndex].value = aState;
}


//
// class methods
//
+ (id) singleInstance
{
  if ( !singleInstance)
    {
      singleInstance = [[AdvancedViewController alloc] initWithNibName: @"AdvancedView"];
    }

  return singleInstance;
}

@end
