/*
**  GNUMail+Extensions.m
**
**  Copyright (c) 2002, 2003
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "Filter.h"
#include "FilterManager.h"
#include "Constants.h"
#include "GNUMail+Extensions.h"
#include "Utilities.h"

#include <Foundation/NSFileManager.h>

#include <Pantomime/URLName.h>

@implementation GNUMail (Extensions)

#define UPDATE_PATH(name) ({ \
 aFolderName = [self updatePathForFolderName: [allValues objectForKey: name] \
			      current: theDestinationPath \
			      previous: theSourcePath]; \
 if ( aFolderName ) \
   { \
     [allValues setObject: aFolderName  forKey: name]; \
   } \
})

//
//
//
- (void) moveLocalMailDirectoryFromPath: (NSString *) theSourcePath
                                 toPath: (NSString *) theDestinationPath
{
  NSArray *foldersToOpen, *theFilters;
  FilterManager *aFilterManager;
  NSFileManager *aFileManager;
  NSAutoreleasePool *pool;
  NSString *aFolderName;

  BOOL aBOOL, isDir;
  
  pool = [[NSAutoreleasePool alloc] init];
  aFileManager = [NSFileManager defaultManager];

  if ( [aFileManager fileExistsAtPath: theDestinationPath
		     isDirectory: &isDir] )
    {
      //
      // The "directory" already exists but it is a file, not a directory.
      //
      if ( !isDir )
	{ 
	  NSRunCriticalAlertPanel(_(@"Fatal error!"),
				  _(@"%@ exists but it is a file not a directory.\nThe application will now terminate.\nRemove this file before trying again to start GNUMail.app."),
				  @"OK",
				  NULL,
				  NULL,
				  theDestinationPath);
	  exit(1);
	}
      //
      // The directory exists, let's move the files there.
      // FIXME: That could be improved. If the destition path already exists,
      //        we will overwrite that mailbox. We should probably ask the
      //        user what he/she wants to do.
      else
	{
	  NSDirectoryEnumerator *aDirectoryEnumerator;
	  NSString *aFile;

	  NSDebugLog(@"The directory exists - we move the files.");

	  aDirectoryEnumerator = [aFileManager enumeratorAtPath: theSourcePath];
	  
	  while ( (aFile = [aDirectoryEnumerator nextObject]) )
	    {
	      [aFileManager movePath: [NSString stringWithFormat: @"%@/%@", theSourcePath, aFile]
			    toPath: [NSString stringWithFormat: @"%@/%@", theDestinationPath, aFile]
			    handler: nil];
	    }

	  // We remove the old directory.
	  [aFileManager removeFileAtPath: theSourcePath  handler: nil];
	}
    }
  //
  // The new target directory doesn't yet exist. We can safely move our current directory
  // to the new one.
  //
  else
    {
      NSDebugLog(@"The directory doesn't exist - we move the directory.");

      aBOOL = [aFileManager movePath: theSourcePath
			    toPath: theDestinationPath
			    handler: nil];
      if ( !aBOOL )
	{
	  NSRunCriticalAlertPanel(_(@"Fatal error!"),
				  _(@"A fatal error occured when moving the directory %@ to %@.\nThe application will now terminate."),
				  @"OK",
				  NULL,
				  NULL,
				  theSourcePath,
				  theDestinationPath);
	  exit(1);
	}
    }
  
  
  //
  // FOLDERS_TO_OPEN
  //
  foldersToOpen = [[NSUserDefaults standardUserDefaults] arrayForKey: @"FOLDERS_TO_OPEN"];
  
  if ( foldersToOpen && [foldersToOpen count] > 0)
    {
      NSMutableArray *aMutableArray;
      int i;
      
      aMutableArray = [NSMutableArray array];
      
      for (i = 0; i < [foldersToOpen count]; i++)
	{
	  aFolderName = [self updatePathForFolderName: [foldersToOpen objectAtIndex: i]
			      current: theDestinationPath
			      previous: theSourcePath];
	  
	  if ( aFolderName )
	    {
	      [aMutableArray addObject: aFolderName];
	    }
	  else
	    {
	      [aMutableArray addObject: [foldersToOpen objectAtIndex: i]];
	    }
	}
      
      [[NSUserDefaults standardUserDefaults] setObject: aMutableArray  forKey: @"FOLDERS_TO_OPEN"];
    }
  
  
  //
  // Filters -> actionFolderName
  //
  aFilterManager = [FilterManager singleInstance];
  theFilters = [aFilterManager filters];
  
  if ( [theFilters count] > 0 )
    {
      Filter *aFilter;
      int i;

      for (i = 0; i < [theFilters count]; i++)
	{
	  aFilter = [theFilters objectAtIndex: i];
	  
	  aFolderName = [self updatePathForFolderName: [aFilter actionFolderName]
			      current: theDestinationPath
			      previous: theSourcePath];
	  
	  if ( aFolderName )
	    {
	      [aFilter setActionFolderName: aFolderName];
	    }
	}
      
      [aFilterManager synchronize];
    }
  
	  
  //
  // ACCOUNTS -> [all accounts] -> RECEIVE -> DRAFTSFOLDERNAME, INBOXFOLDERNAME, SENTFOLDERNAME, TRASHFOLDERNAME
  //
  if ( [[NSUserDefaults standardUserDefaults] objectForKey: @"ACCOUNTS"] )
    {
      NSMutableDictionary *allAccounts, *theAccount, *allValues;
      NSEnumerator *theEnumerator;
      NSString *aKey;
      
      allAccounts = [[NSMutableDictionary alloc] initWithDictionary: [[NSUserDefaults standardUserDefaults] 
								       dictionaryForKey: @"ACCOUNTS"]];
      
      theEnumerator = [allAccounts keyEnumerator];

      while ( (aKey = [theEnumerator nextObject]) )
	{
	  theAccount = [[NSMutableDictionary alloc] initWithDictionary: [allAccounts objectForKey: aKey]];
	  allValues = [[NSMutableDictionary alloc] initWithDictionary: [theAccount objectForKey: @"MAILBOXES"]];

	  UPDATE_PATH(@"DRAFTSFOLDERNAME");
	  UPDATE_PATH(@"INBOXFOLDERNAME");
	  UPDATE_PATH(@"SENTFOLDERNAME");
	  UPDATE_PATH(@"TRASHFOLDERNAME");

	  [theAccount setObject: allValues  forKey: @"MAILBOXES"];
	  RELEASE(allValues);

	  [allAccounts setObject: theAccount  forKey: aKey];
	  RELEASE(theAccount);
	}
      
      [[NSUserDefaults standardUserDefaults] setObject: allAccounts  forKey: @"ACCOUNTS"];
      RELEASE(allAccounts);
    }
  
  
  // We set the path of our previous default's value and we synchronize our defaults.
  [[NSUserDefaults standardUserDefaults] setObject: theDestinationPath  forKey: @"LOCALMAILDIR_PREVIOUS"];
  [[NSUserDefaults standardUserDefaults] synchronize];
  
  RELEASE(pool);
}


//
// In 1.0.99, we must get the previous PERSONAL preferences and best guess
// the PERSONAL preference of an ACCOUNT based on the RECEIVING key - this
// method does that.
//
- (NSDictionary *) personalPreferencesUsingReceivingKey: (NSString *) theKey
{
  NSDictionary *allPersonals;

  allPersonals = [[NSUserDefaults standardUserDefaults] objectForKey: @"PERSONAL"];

  if ( !allPersonals || [allPersonals count] == 0 )
    {
      return [NSDictionary dictionary];
    }
  else if ( [allPersonals count] == 1 )
    {
      return [[allPersonals objectEnumerator] nextObject];
    }
  else
    {
      NSString *theEMailAddress, *aString;
      NSArray *allKeys;
      
      NSRange aRange;
      int i;

      allKeys = [allPersonals allKeys];

      for (i = 0; i < [allKeys count]; i++)
	{
	  
	  theEMailAddress = [[allPersonals objectForKey: [allKeys objectAtIndex: i]] objectForKey: @"EMAILADDR"];
	  aRange = [theEMailAddress rangeOfString: @"@"];
	  
	  if ( aRange.length )
	    {
	      aString = [theEMailAddress substringFromIndex: aRange.location+1];

	      aRange = [theKey rangeOfString: aString];

	      // If the E-Mail address "matches" our receiving host, we use that profile
	      if ( aRange.length )
		{
		  return [allPersonals objectForKey: [allKeys objectAtIndex: i]];
		}
	    }
	}

      // If we haven't found anything, we return the first one.
      return [allPersonals objectForKey: [allKeys objectAtIndex: 0]];
    }
}


//
// This method is used to remove deprecated preferences values
//
- (void) removeDeprecatedDefaults
{
  // In 1.0.98, this is no longer used.
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"SHOWICONS"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"SHOWWAITPANELS"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"RECT_EDITWINDOW"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"RECT_MAILWINDOW"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"SIGNATURE"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"SIGNATURE_SOURCE"];
  
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"TAG_FONT_NAME"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"TAG_FONT_SIZE"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"HEADER_FONT_NAME"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"HEADER_FONT_SIZE"];

  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"MAILSPOOLFILE"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"TRANSFERFROMMAILSPOOLFILE"];

  // In 1.0.99, this is no longer used.
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"DRAFTSFOLDERNAME"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"INBOXFOLDERNAME"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"OUTBOXFOLDERNAME"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"PERSONAL"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"RETRIEVEMETHOD"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"RETRIEVEMINUTES"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"REUSE_MAILWINDOW"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"SENDING"];
  [[NSUserDefaults standardUserDefaults] removeObjectForKey: @"TRASHFOLDERNAME"];
}


//
//
//
- (void) removeTemporaryFiles
{
  NSDirectoryEnumerator *theEnumerator;
  NSString *aPath;

  theEnumerator = [[NSFileManager defaultManager] enumeratorAtPath: GNUMailTemporaryDirectory()];

  while ( (aPath = [theEnumerator nextObject]) )
    {
      [[NSFileManager defaultManager] removeFileAtPath: [NSString stringWithFormat: @"%@/%@", GNUMailTemporaryDirectory(), aPath] 
				      handler: nil];
    }
}


//
// In 1.0.99, we must get the previous SENDING preferences and best guess
// SEND preferences of an ACCOUNT based on the RECEIVING key - this method does that.
//
- (NSDictionary *) sendPreferencesUsingReceivingKey: (NSString *) theKey
{
  NSArray *allServers;

  allServers = [[NSUserDefaults standardUserDefaults] objectForKey: @"SENDING"];

  if ( !allServers || ![allServers count] )
    {
      return [NSDictionary dictionary];
    }
  else if ([allServers count] == 1)
    {
      return [allServers objectAtIndex: 0];
    }
  else
    {
      NSString *aSendingName, *aReceivingName;
      int i;
      
      aReceivingName = [[[[NSUserDefaults standardUserDefaults] objectForKey: @"RECEIVING"]
			  objectForKey: theKey] objectForKey: @"SERVERNAME"];
      
      for (i = 0; i < [allServers count]; i++)
	{
	  aSendingName = [[allServers objectAtIndex: i] objectForKey: @"SMTP_HOST"];
	  
	  // We verify if SMTP_HOST is equal to SERVERNAME
	  if ( aSendingName && 
	       aReceivingName &&
	       [[aSendingName uppercaseString] isEqualToString: [aReceivingName uppercaseString]] )
	    {
	      return [allServers objectAtIndex: i];
	    }
	}
      
      // If we haven't found anything, we return the first one.
      return [allServers objectAtIndex: 0];
    }

  // Never reached
  return [NSDictionary dictionary];
}


//
// From 1.0.98 (1.1.0pre1) -> 1.0.99 (1.1.0pre2)
//
- (void) update_1098_to_1099
{
  NSUserDefaults *aStandardUserDefaults;
  int i;

  aStandardUserDefaults = [NSUserDefaults standardUserDefaults];
  i = 1;

  if ( [aStandardUserDefaults objectForKey: @"RECEIVING"] )
    {
      NSMutableDictionary *allServers, *allAccounts, *theAccount;
      NSEnumerator *theEnumerator;
      NSString *aKey;

      allAccounts = [[NSMutableDictionary alloc] init];
      
      allServers = [aStandardUserDefaults objectForKey: @"RECEIVING"];
      theEnumerator = [allServers keyEnumerator];

      while ( (aKey = [theEnumerator nextObject]) )
	{
	  NSMutableDictionary *allValues;

       	  theAccount = [[NSMutableDictionary alloc] init];

	  //
	  // We migrate the RECEIVING preferences.
	  //
	  allValues = [[NSMutableDictionary alloc] init];
	  [allValues addEntriesFromDictionary: [allServers objectForKey: aKey]];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"RETRIEVEMETHOD"]  forKey: @"RETRIEVEMETHOD"];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"RETRIEVEMINUTES"]  forKey: @"RETRIEVEMINUTES"];
	  [theAccount setObject: allValues  forKey: @"RECEIVE"];
	  RELEASE(allValues);

	  //
	  // We migrate the PERSONAL preferences.
	  //
	  allValues = [[NSMutableDictionary alloc] init];
	  [allValues addEntriesFromDictionary: [self personalPreferencesUsingReceivingKey: aKey]];
	  [theAccount setObject: allValues  forKey: @"PERSONAL"];
	  RELEASE(allValues);

	  //
	  // We migrate the SENDING preferences.
	  //
	  allValues = [[NSMutableDictionary alloc] init];
	  [allValues addEntriesFromDictionary: [self sendPreferencesUsingReceivingKey: aKey]];
	  [allValues removeObjectForKey: @"PERSONAL_PROFILE"];
	  [theAccount setObject: allValues  forKey: @"SEND"];
	  RELEASE(allValues);
	  
	  //
	  // We migrate the MAILBOXES preferences.
	  //
	  allValues = [[NSMutableDictionary alloc] init];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"DRAFTSFOLDERNAME"]  forKey: @"DRAFTSFOLDERNAME"];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"INBOXFOLDERNAME"]   forKey: @"INBOXFOLDERNAME"];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"OUTBOXFOLDERNAME"]  forKey: @"OUTBOXFOLDERNAME"];
	  [allValues setObject: [aStandardUserDefaults objectForKey: @"TRASHFOLDERNAME"]   forKey: @"TRASHFOLDERNAME"];
	  [theAccount setObject: allValues  forKey: @"MAILBOXES"];
	  RELEASE(allValues);

	  //
	  // We finally add this new account, as enabled. The account name is automatically generated
	  // with the "Account x" name.
	  //
	  [theAccount setObject: [NSNumber numberWithBool: YES]  forKey: @"ENABLED"];
	  [theAccount setObject: [NSNumber numberWithBool: NO]  forKey: @"DEFAULT"];
	  [allAccounts setObject: theAccount  forKey: [NSString stringWithFormat: @"Account %d", i]];
	  i++;

	  RELEASE(theAccount);
	}
     
      
      [aStandardUserDefaults setObject: allAccounts  forKey: @"ACCOUNTS"];
      [aStandardUserDefaults removeObjectForKey: @"RECEIVING"];
      RELEASE(allAccounts);

      // We warn the user about this important "migration" of the preferences.
      NSRunInformationalAlertPanel(_(@"Important note!"),
				   _(@"In 1.1.0pre2, the preferences have changed considerably. The old preferences\nhave been migrated but you should verify all your accounts in\nPreferences > Account to be sure that the information is valid.\nSorry for this inconvenience."),
				   _(@"OK"),
				   NULL, 
				   NULL,
				   NULL);
    } 
}


//
// From 1.0.99 (1.1.0pre2) -> 1.1.0 final.
//
- (void) update_1099_to_110
{
  NSUserDefaults *aStandardUserDefaults;

  aStandardUserDefaults = [NSUserDefaults standardUserDefaults];
  
  if ( [aStandardUserDefaults objectForKey: @"ACCOUNTS"] )
    {
      NSMutableDictionary *allAccounts, *allValues, *allMailboxValues;
      NSEnumerator *theEnumerator;
      NSString *aKey;

      allAccounts = [[NSMutableDictionary alloc] initWithDictionary: [aStandardUserDefaults objectForKey: @"ACCOUNTS"]];
      
      theEnumerator = [allAccounts keyEnumerator];
      
      while ( (aKey = [theEnumerator nextObject]) )
	{
	  allValues = [[NSMutableDictionary alloc] initWithDictionary: [allAccounts objectForKey: aKey]];
	  allMailboxValues = [[NSMutableDictionary alloc] initWithDictionary: [allValues objectForKey: @"MAILBOXES"]];

	  // We replace OUTBOXFOLDERNAME by SENTFOLDERNAME, if we need to.
	  if ( [allMailboxValues objectForKey: @"OUTBOXFOLDERNAME"] )
	    {
	      [allMailboxValues setObject: [allMailboxValues objectForKey: @"OUTBOXFOLDERNAME"]  forKey: @"SENTFOLDERNAME"];
	      [allMailboxValues removeObjectForKey: @"OUTBOXFOLDERNAME"];
	    }

	  [allValues setObject: allMailboxValues  forKey: @"MAILBOXES"];
	  RELEASE(allMailboxValues);

	  [allAccounts setObject: allValues  forKey: aKey];
	  RELEASE(allValues);
	}


      [aStandardUserDefaults setObject: allAccounts  forKey: @"ACCOUNTS"];
      RELEASE(allAccounts);
    }
}


//
//
//
- (NSString *) updatePathForFolderName: (NSString *) theFolderName
			       current: (NSString *) theCurrentPath
			      previous: (NSString *) thePreviousPath
{
  URLName *aURLName;

  if ( !theFolderName || 
       ![theFolderName hasPrefix: [NSString stringWithFormat: @"local://%@", thePreviousPath]] )
    {
      return nil;
    }
  
  aURLName = [[URLName alloc] initWithString: theFolderName  path: thePreviousPath];
  AUTORELEASE(aURLName);

  return [NSString stringWithFormat: @"local://%@/%@", theCurrentPath, [aURLName foldername]];
}

@end

#undef UPDATE_PATH
