/*
**  Group.m
**
**  Copyright (c) 2001, 2002, 2003
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "Group.h"

#include "Address.h"
#include "Constants.h"

#ifdef MACOSX
#import <AddressBook/AddressBook.h>
#endif


@implementation Group

//
//
//
- (id) init
{
  self = [super init];

  addresses = [[NSMutableArray alloc] init];
  
  return self;
}


//
//
//
- (id) initWithName: (NSString *) theName
{
  self = [self init];

  [self setName: theName];

  return self;
}


//
//
//
- (void) dealloc
{
  RELEASE(name);
  RELEASE(addresses);

  [super dealloc];
}


//
// NSCoding protocol
//
- (void) encodeWithCoder: (NSCoder *) theCoder
{
  [theCoder encodeObject: [self name] ];
  [theCoder encodeObject: [self addresses]];
}

- (id) initWithCoder: (NSCoder *) theCoder
{
  self = [super init];
  
  addresses = [[NSMutableArray alloc] init];

  [self setName: [theCoder decodeObject]];
  [self setAddresses: [theCoder decodeObject] ];

  return self;
}


//
// access/mutation methods
//
- (NSString *) name
{
  return name;
}


//
//
//
- (void) setName: (NSString *) theName
{
  RETAIN(theName);
  RELEASE(name);
  name = theName;
}


//
//
//
- (Address *) addressAtIndex: (int) theIndex
{
  return [addresses objectAtIndex: theIndex];
}


//
//
//
- (NSArray *) addresses
{
#ifdef MACOSX  // Use Apple's addressbook if desired; default operation
  if ( ([[NSUserDefaults standardUserDefaults] stringForKey: @"ADDRESSBOOK_TYPE"] == nil) ||
       ([[NSUserDefaults standardUserDefaults] integerForKey: @"ADDRESSBOOK_TYPE"] == APPLE_ADDRESSBOOK_TYPE) )
    {
      ABAddressBook *AB = [ABAddressBook sharedAddressBook];
      ABSearchElement *searchGroup;
      ABGroup *thisABGroup, *aSubgroup;
      ABPerson *thisABPerson;
      NSEnumerator *abGroupEnumerator, *memberEnumerator, *subgroupEnumerator;
      NSArray *theGroups;
      NSMutableArray *memberArray;
      
      searchGroup = [ABGroup searchElementForProperty:kABGroupNameProperty
			     label:nil
			     key:nil
			     value:[self name]
			     comparison:kABEqual];
      
      
      theGroups = [AB recordsMatchingSearchElement: searchGroup];
      
      abGroupEnumerator = [theGroups objectEnumerator];
      memberArray = [[NSMutableArray alloc] initWithCapacity: 1];
      
      while ( (thisABGroup = [abGroupEnumerator nextObject]) )
        {
	  memberEnumerator = [[thisABGroup members] objectEnumerator];
	  subgroupEnumerator = [[thisABGroup subgroups] objectEnumerator];
	  
	  while( (thisABPerson = [memberEnumerator nextObject]) )
            {
	      Address *theAddress;
	      int emailCount, i;
	      
	      emailCount = [[thisABPerson valueForProperty: kABEmailProperty] count];
	      for (i = 0; i < emailCount; i++)
                {
		  theAddress = [[Address alloc] init];
		  [theAddress setName: [NSString stringWithFormat:@"%@ %@", 
						 [thisABPerson valueForProperty: kABFirstNameProperty],
						 [thisABPerson valueForProperty: kABLastNameProperty]]]; 
		  [theAddress setEmailAddress: [[thisABPerson valueForProperty: kABEmailProperty] valueAtIndex: i]];
		  [memberArray addObject: theAddress];  
		  RELEASE(theAddress);
                }        
	      
            }
	  
	  // recurse through subgroups and get all the members
	  while ( (aSubgroup = [subgroupEnumerator nextObject]) )
            {
	      Group *tmpGroup;
	      
	      tmpGroup = [[Group alloc] init];
	      [tmpGroup setName: [aSubgroup valueForProperty: kABGroupNameProperty]];
	      [memberArray addObjectsFromArray: [tmpGroup addresses]];
	      AUTORELEASE(tmpGroup);
            }            
	  
        }
      
        return (AUTORELEASE(memberArray));
    }
#endif
  
  // GNUMail.app addressbook type
  return [NSArray arrayWithArray: addresses];
}


//
//
//
- (void) setAddresses: (NSArray *) theAddresses
{
  [addresses addObjectsFromArray: theAddresses]; 
}


//
// This method will verify if an address is already in the
// group before adding it.
//
- (void) addAddress: (Address *) theAddress
{
  if ( theAddress )
    {
      int i;
      
      for (i = 0; i < [addresses count]; i++)
	{
	  Address *anAddress;

	  anAddress = [addresses objectAtIndex: i];
	  
	  if ( [[anAddress description] caseInsensitiveCompare: [theAddress description]] == NSOrderedSame)
	    {
	      return;
	    }
	}
      
      [addresses addObject: theAddress];
    }
  else
    {
      NSDebugLog(@"Tried to add a nil object to the group %@", [self name]);
    }
}

- (void) removeAddress: (Address *) theAddress
{
  if ( theAddress )
    {
      [addresses removeObject: theAddress];
    }
  else
    {
      NSDebugLog(@"Tried to remove a nil object from the group %@", [self name]);
    }
}

- (void) removeAllAddresses
{
  [addresses removeAllObjects];
}

- (NSComparisonResult) compareAccordingToName: (Group *) aGroup
{
  return [name compare: [aGroup name]];
}

- (void) sort
{
  [addresses sortUsingSelector: @selector(compareAccordingToName:)];
}

- (NSString *) description
{
  return [self name];
}

@end
